var ref = require('ref'); // DLL Data types 
var ffi = require('ffi'); // DLL interface 

CASBACNETSTACK_LIBARY_ARCHITECTURE = 'x64'; // x64 or x86 
CASBACNETSTACK_LIBARY_TYPE = 'Debug'; // Release or Debug
CASBACNETSTACK_LIBARY_PATH = './bin/';

// Select the right shared libary for the OS. 
var CASBACNETSTACK_LIBARY_FILENAME
if (process.platform == "win32") {
    CASBACNETSTACK_LIBARY_FILENAME = CASBACNETSTACK_LIBARY_PATH + 'CASBACnetStack_' + CASBACNETSTACK_LIBARY_ARCHITECTURE + '_' + CASBACNETSTACK_LIBARY_TYPE + '.dll'
} else {
    CASBACNETSTACK_LIBARY_FILENAME = CASBACNETSTACK_LIBARY_PATH + 'libCASBACnetStack_' + CASBACNETSTACK_LIBARY_ARCHITECTURE + '_' + CASBACNETSTACK_LIBARY_TYPE
}

module.exports = {
    GetAdapterVersion: function GetAdapterVersion() {
        return "1.0.0.0";
    },
    GetLibaryPath: function GetLibaryPath() {
        return (CASBACNETSTACK_LIBARY_FILENAME);
    },
    stack: ffi.Library(CASBACNETSTACK_LIBARY_FILENAME, {
        'BACnetStack_GetAPIMajorVersion': ['uint32', []],
        'BACnetStack_GetAPIMinorVersion': ['uint32', []],
        'BACnetStack_GetAPIPatchVersion': ['uint32', []],
        'BACnetStack_GetAPIBuildVersion': ['uint32', []],

        'BACnetStack_Loop': ['void', []],

        'BACnetStack_RegisterCallbackSendMessage': ['uint16', ['pointer']],
        'BACnetStack_RegisterCallbackReceiveMessage': ['uint16', ['pointer']],
        'BACnetStack_RegisterCallbackGetSystemTime': ['void', ['pointer']],

        'BACnetStack_AddDevice': ['bool', ['uint32']],

        // bool BACnetStack_AddObject(const uint32_t deviceInstance, const uint16_t objectType, const uint32_t objectInstance)
        'BACnetStack_AddObject': ['bool', ['uint32', 'uint16', 'uint32',]],

        // GetProperty
        'BACnetStack_RegisterCallbackGetPropertyCharacterString': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyReal': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyBool': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyEnumerated': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyDate': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyDouble': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyOctetString': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertySignedInteger': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyTime': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackGetPropertyUnsignedInteger': ['void', ['pointer']],

        // Set Propery 
        'BACnetStack_RegisterCallbackSetPropertyReal': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyEnumerated': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyBitString': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyBool': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyCharacterString': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyDate': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyDouble': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyNull': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyOctetString': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertySignedInteger': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyTime': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetPropertyUnsignedInteger': ['void', ['pointer']],
        'BACnetStack_RegisterCallbackSetElevatorGroupLandingCallControl': ['void', ['pointer']],

        // 
        'BACnetStack_SetServiceEnabled': ['void', ['uint32', 'uint32', 'bool']],
        'BACnetStack_SetPropertyWritable': ['void', ['uint32', 'uint16', 'uint32', 'uint32', 'bool']],
        'BACnetStack_SetPropertyEnabled': ['void', ['uint32', 'uint16', 'uint32', 'uint32', 'bool']],
        'BACnetStack_SetPropertyByObjectTypeWritable': ['void', ['uint32', 'uint16', 'uint32', 'uint32', 'bool']],
        'BACnetStack_SetPropertyByObjectTypeEnabled': ['void', ['uint32', 'uint16', 'uint32', 'uint32', 'bool']],
        'BACnetStack_ValueUpdated': ['void', ['uint32', 'uint16', 'uint32', 'uint32']],


    }),

    SERVICES_SUPPORTED: {
        ACKNOWLEDGE_ALARM: 0,
        CONFIRMED_COV_NOTIFICATION: 1,
        CONFIRMED_COV_NOTIFICATION_MULTIPLE: 42,
        CONFIRMED_EVENT_NOTIFICATION: 2,
        GET_ALARM_SUMMARY: 3,
        GET_ENROLLMENT_SUMMARY: 4,
        GET_EVENT_INFORMATION: 39,
        LIFE_SAFETY_OPERATION: 37,
        SUBSCRIBE_COV: 5,
        SUBSCRIBE_COV_PROPERTY: 38,
        SUBSCRIBE_COV_PROPERTY_MULTIPLE: 41,
        ATOMIC_READ_FILE: 6,
        ATOMIC_WRITE_FILE: 7,
        ADD_LIST_ELEMENT: 8,
        REMOVE_LIST_ELEMENT: 9,
        CREATE_OBJECT: 10,
        DELETE_OBJECT: 11,
        READ_PROPERTY: 12,
        READ_PROPERTY_MULTIPLE: 14,
        READ_RANGE: 35,
        WRITE_GROUP: 40,
        WRITE_PROPERTY: 15,
        WRITE_PROPERTY_MULTIPLE: 16,
        DEVICE_COMMUNICATION_CONTROL: 17,
        CONFIRMED_PRIVATE_TRANSFER: 18,
        CONFIRMED_TEXT_MESSAGE: 19,
        REINITIALIZE_DEVICE: 20,
        VT_OPEN: 21,
        VT_CLOSE: 22,
        VT_DATA: 23,
        I_AM: 26,
        I_HAVE: 27,
        UNCONFIRMED_COV_NOTIFICATION: 28,
        UNCONFIRMED_COV_NOTIFICATION_MULTIPLE: 43,
        UNCONFIRMED_EVENT_NOTIFICATION: 29,
        UNCONFIRMED_PRIVATE_TRANSFER: 30,
        UNCONFIRMED_TEXT_MESSAGE: 31,
        TIME_SYNCHRONIZATION: 32,
        UTC_TIME_SYNCHRONIZATION: 36,
        WHO_HAS: 33,
        WHO_IS: 34,
    },
    OBJECT_TYPE: {
        ANALOG_INPUT: 0,
        ANALOG_OUTPUT: 1,
        ANALOG_VALUE: 2,
        BINARY_INPUT: 3,
        BINARY_OUTPUT: 4,
        BINARY_VALUE: 5,
        CALENDAR: 6,
        COMMAND: 7,
        DEVICE: 8,
        EVENT_ENROLLMENT: 9,
        FILE: 10,
        GROUP: 11,
        LOOP: 12,
        MULTI_STATE_INPUT: 13,
        MULTI_STATE_OUTPUT: 14,
        NOTIFICATION_CLASS: 15,
        PROGRAM: 16,
        SCHEDULE: 17,
        AVERAGING: 18,
        MULTI_STATE_VALUE: 19,
        TREND_LOG: 20,
        LIFE_SAFETY_POINT: 21,
        LIFE_SAFETY_ZONE: 22,
        ACCUMULATOR: 23,
        PULSE_CONVERTER: 24,
        EVENT_LOG: 25,
        GLOBAL_GROUP: 26,
        TREND_LOG_MULTIPLE: 27,
        LOAD_CONTROL: 28,
        STRUCTURED_VIEW: 29,
        ACCESS_DOOR: 30,
        TIMER: 31,
        ACCESS_CREDENTIAL: 32,
        ACCESS_POINT: 33,
        ACCESS_RIGHTS: 34,
        ACCESS_USER: 35,
        ACCESS_ZONE: 36,
        CREDENTIAL_DATA_INPUT: 37,
        NETWORK_SECURITY: 38,
        BITSTRING_VALUE: 39,
        CHARACTERSTRING_VALUE: 40,
        DATEPATTERN_VALUE: 41,
        DATE_VALUE: 42,
        DATETIMEPATTERN_VALUE: 43,
        DATETIME_VALUE: 44,
        INTEGER_VALUE: 45,
        LARGE_ANALOG_VALUE: 46,
        OCTETSTRING_VALUE: 47,
        POSITIVE_INTEGER_VALUE: 48,
        TIMEPATTERN_VALUE: 49,
        TIME_VALUE: 50,
        NOTIFICATION_FORWARDER: 51,
        ALERT_ENROLLMENT: 52,
        CHANNEL: 53,
        LIGHTING_OUTPUT: 54,
        BINARY_LIGHTING_OUTPUT: 55,
        NETWORK_PORT: 56,
        ELEVATOR_GROUP: 57,
        ESCALATOR: 58,
        LIFT: 59
    },

    PROPERTY_IDENTIFIER: {
        ABSENTEELIMIT: 244,
        ACCEPTEDMODES: 175,
        ACCESSALARMEVENTS: 245,
        ACCESSDOORS: 246,
        ACCESSEVENT: 247,
        ACCESSEVENTAUTHENTICATIONFACTOR: 248,
        ACCESSEVENTCREDENTIAL: 249,
        ACCESSEVENTTAG: 322,
        ACCESSEVENTTIME: 250,
        ACCESSTRANSACTIONEVENTS: 251,
        ACCOMPANIMENT: 252,
        ACCOMPANIMENTTIME: 253,
        ACKREQUIRED: 1,
        ACKEDTRANSITIONS: 0,
        ACTION: 2,
        ACTIONTEXT: 3,
        ACTIVATIONTIME: 254,
        ACTIVEAUTHENTICATIONPOLICY: 255,
        ACTIVECOVMULTIPLESUBSCRIPTIONS: 481,
        ACTIVECOVSUBSCRIPTIONS: 152,
        ACTIVETEXT: 4,
        ACTIVEVTSESSIONS: 5,
        ACTUALSHEDLEVEL: 212,
        ADJUSTVALUE: 176,
        ALARMVALUE: 6,
        ALARMVALUES: 7,
        ALIGNINTERVALS: 193,
        ALL: 8,
        ALLWRITESSUCCESSFUL: 9,
        ALLOWGROUPDELAYINHIBIT: 365,
        APDULENGTH: 399,
        APDUSEGMENTTIMEOUT: 10,
        APDUTIMEOUT: 11,
        APPLICATIONSOFTWAREVERSION: 12,
        ARCHIVE: 13,
        ASSIGNEDACCESSRIGHTS: 256,
        ASSIGNEDLANDINGCALLS: 447,
        ATTEMPTEDSAMPLES: 124,
        AUTHENTICATIONFACTORS: 257,
        AUTHENTICATIONPOLICYLIST: 258,
        AUTHENTICATIONPOLICYNAMES: 259,
        AUTHENTICATIONSTATUS: 260,
        AUTHORIZATIONEXEMPTIONS: 364,
        AUTHORIZATIONMODE: 261,
        AUTOSLAVEDISCOVERY: 169,
        AVERAGEVALUE: 125,
        BACKUPANDRESTORESTATE: 338,
        BACKUPFAILURETIMEOUT: 153,
        BACKUPPREPARATIONTIME: 339,
        BACNETIPGLOBALADDRESS: 407,
        BACNETIPMODE: 408,
        BACNETIPMULTICASTADDRESS: 409,
        BACNETIPNATTRAVERSAL: 410,
        BACNETIPUDPPORT: 412,
        BACNETIPV6MODE: 435,
        BACNETIPV6UDPPORT: 438,
        BACNETIPV6MULTICASTADDRESS: 440,
        BASEDEVICESECURITYPOLICY: 327,
        BBMDACCEPTFDREGISTRATIONS: 413,
        BBMDBROADCASTDISTRIBUTIONTABLE: 414,
        BBMDFOREIGNDEVICETABLE: 415,
        BELONGSTO: 262,
        BIAS: 14,
        BITMASK: 342,
        BITTEXT: 343,
        BLINKWARNENABLE: 373,
        BUFFERSIZE: 126,
        CARASSIGNEDDIRECTION: 448,
        CARDOORCOMMAND: 449,
        CARDOORSTATUS: 450,
        CARDOORTEXT: 451,
        CARDOORZONE: 452,
        CARDRIVESTATUS: 453,
        CARLOAD: 454,
        CARLOADUNITS: 455,
        CARMODE: 456,
        CARMOVINGDIRECTION: 457,
        CARPOSITION: 458,
        CHANGEOFSTATECOUNT: 15,
        CHANGEOFSTATETIME: 16,
        CHANGESPENDING: 416,
        CHANNELNUMBER: 366,
        CLIENTCOVINCREMENT: 127,
        COMMAND: 417,
        COMMANDTIMEARRAY: 430,
        CONFIGURATIONFILES: 154,
        CONTROLGROUPS: 367,
        CONTROLLEDVARIABLEREFERENCE: 19,
        CONTROLLEDVARIABLEUNITS: 20,
        CONTROLLEDVARIABLEVALUE: 21,
        COUNT: 177,
        COUNTBEFORECHANGE: 178,
        COUNTCHANGETIME: 179,
        COVINCREMENT: 22,
        COVPERIOD: 180,
        COVRESUBSCRIPTIONINTERVAL: 128,
        COVUPERIOD: 349,
        COVURECIPIENTS: 350,
        CREDENTIALDISABLE: 263,
        CREDENTIALSTATUS: 264,
        CREDENTIALS: 265,
        CREDENTIALSINZONE: 266,
        CURRENTCOMMANDPRIORITY: 431,
        DATABASEREVISION: 155,
        DATELIST: 23,
        DAY_LIGHT_SAVINGS_STATUS: 24,
        DAYSREMAINING: 267,
        DEADBAND: 25,
        DEFAULTFADETIME: 374,
        DEFAULTRAMPRATE: 375,
        DEFAULTSTEPINCREMENT: 376,
        DEFAULTSUBORDINATERELATIONSHIP: 490,
        DEFAULTTIMEOUT: 393,
        DEPLOYEDPROFILELOCATION: 484,
        DERIVATIVECONSTANT: 26,
        DERIVATIVECONSTANTUNITS: 27,
        DESCRIPTION: 28,
        DESCRIPTIONOFHALT: 29,
        DEVICEADDRESSBINDING: 30,
        DEVICETYPE: 31,
        DIRECTREADING: 156,
        DISTRIBUTIONKEYREVISION: 328,
        DONOTHIDE: 329,
        DOORALARMSTATE: 226,
        DOOREXTENDEDPULSETIME: 227,
        DOORMEMBERS: 228,
        DOOROPENTOOLONGTIME: 229,
        DOORPULSETIME: 230,
        DOORSTATUS: 231,
        DOORUNLOCKDELAYTIME: 232,
        DUTYWINDOW: 213,
        EFFECTIVEPERIOD: 32,
        EGRESSACTIVE: 386,
        EGRESSTIME: 377,
        ELAPSEDACTIVETIME: 33,
        ELEVATORGROUP: 459,
        ENABLE: 133,
        ENERGYMETER: 460,
        ENERGYMETERREF: 461,
        ENTRYPOINTS: 268,
        ERRORLIMIT: 34,
        ESCALATORMODE: 462,
        EVENTALGORITHMINHIBIT: 354,
        EVENTALGORITHMINHIBITREF: 355,
        EVENTDETECTIONENABLE: 353,
        EVENTENABLE: 35,
        EVENTMESSAGETEXTS: 351,
        EVENTMESSAGETEXTSCONFIG: 352,
        EVENTPARAMETERS: 83,
        EVENTSTATE: 36,
        EVENTTIMESTAMPS: 130,
        EVENTTYPE: 37,
        EXCEPTIONSCHEDULE: 38,
        EXECUTIONDELAY: 368,
        EXITPOINTS: 269,
        EXPECTEDSHEDLEVEL: 214,
        EXPIRATIONTIME: 270,
        EXTENDEDTIMEENABLE: 271,
        FAILEDATTEMPTEVENTS: 272,
        FAILEDATTEMPTS: 273,
        FAILEDATTEMPTSTIME: 274,
        FAULTHIGHLIMIT: 388,
        FAULTLOWLIMIT: 389,
        FAULTPARAMETERS: 358,
        FAULTSIGNALS: 463,
        FAULTTYPE: 359,
        FAULTVALUES: 39,
        FDBBMDADDRESS: 418,
        FDSUBSCRIPTIONLIFETIME: 419,
        FEEDBACKVALUE: 40,
        FILEACCESSMETHOD: 41,
        FILESIZE: 42,
        FILETYPE: 43,
        FIRMWAREREVISION: 44,
        FLOORTEXT: 464,
        FULLDUTYBASELINE: 215,
        GLOBALIDENTIFIER: 323,
        GROUPID: 465,
        GROUPMEMBERNAMES: 346,
        GROUPMEMBERS: 345,
        GROUPMODE: 467,
        HIGHLIMIT: 45,
        HIGHERDECK: 468,
        INPROCESS: 47,
        INPROGRESS: 378,
        INACTIVETEXT: 46,
        INITIALTIMEOUT: 394,
        INPUTREFERENCE: 181,
        INSTALLATIONID: 469,
        INSTANCEOF: 48,
        INSTANTANEOUSPOWER: 379,
        INTEGRALCONSTANT: 49,
        INTEGRALCONSTANTUNITS: 50,
        INTERFACEVALUE: 387,
        INTERVALOFFSET: 195,
        IPADDRESS: 400,
        IPDEFAULTGATEWAY: 401,
        IPDHCPENABLE: 402,
        IPDHCPLEASETIME: 403,
        IPDHCPLEASETIMEREMAINING: 404,
        IPDHCPSERVER: 405,
        IPDNSSERVER: 406,
        IPSUBNETMASK: 411,
        IPV6ADDRESS: 436,
        IPV6AUTOADDRESSINGENABLE: 442,
        IPV6DEFAULTGATEWAY: 439,
        IPV6DHCPLEASETIME: 443,
        IPV6DHCPLEASETIMEREMAINING: 444,
        IPV6DHCPSERVER: 445,
        IPV6DNSSERVER: 441,
        IPV6PREFIXLENGTH: 437,
        IPV6ZONEINDEX: 446,
        ISUTC: 344,
        KEYSETS: 330,
        LANDINGCALLCONTROL: 471,
        LANDINGCALLS: 470,
        LANDINGDOORSTATUS: 472,
        LASTACCESSEVENT: 275,
        LASTACCESSPOINT: 276,
        LASTCOMMANDTIME: 432,
        LASTCREDENTIALADDED: 277,
        LASTCREDENTIALADDEDTIME: 278,
        LASTCREDENTIALREMOVED: 279,
        LASTCREDENTIALREMOVEDTIME: 280,
        LASTKEYSERVER: 331,
        LASTNOTIFYRECORD: 173,
        LASTPRIORITY: 369,
        LASTRESTARTREASON: 196,
        LASTRESTORETIME: 157,
        LASTSTATECHANGE: 395,
        LASTUSETIME: 281,
        LIFESAFETYALARMVALUES: 166,
        LIGHTINGCOMMAND: 380,
        LIGHTINGCOMMANDDEFAULTPRIORITY: 381,
        LIMITENABLE: 52,
        LIMITMONITORINGINTERVAL: 182,
        LINKSPEED: 420,
        LINKSPEEDAUTONEGOTIATE: 422,
        LINKSPEEDS: 421,
        LISTOFGROUPMEMBERS: 53,
        LISTOFOBJECTPROPERTYREFERENCES: 54,
        LOCAL_DATE: 56,
        LOCALFORWARDINGONLY: 360,
        LOCAL_TIME: 57,
        LOCATION: 58,
        LOCKSTATUS: 233,
        LOCKOUT: 282,
        LOCKOUTRELINQUISHTIME: 283,
        LOGBUFFER: 131,
        LOGDEVICEOBJECTPROPERTY: 132,
        LOGINTERVAL: 134,
        LOGGINGOBJECT: 183,
        LOGGINGRECORD: 184,
        LOGGINGTYPE: 197,
        LOWDIFFLIMIT: 390,
        LOWLIMIT: 59,
        LOWERDECK: 473,
        MACADDRESS: 423,
        MACHINEROOMID: 474,
        MAINTENANCEREQUIRED: 158,
        MAKINGCARCALL: 475,
        MANIPULATEDVARIABLEREFERENCE: 60,
        MANUALSLAVEADDRESSBINDING: 170,
        MASKEDALARMVALUES: 234,
        MAXACTUALVALUE: 382,
        MAXAPDULENGTHACCEPTED: 62,
        MAXFAILEDATTEMPTS: 285,
        MAXINFOFRAMES: 63,
        MAXMASTER: 64,
        MAXPRESVALUE: 65,
        MAXSEGMENTSACCEPTED: 167,
        MAXIMUMOUTPUT: 61,
        MAXIMUMVALUE: 135,
        MAXIMUMVALUETIMESTAMP: 149,
        MEMBEROF: 159,
        MEMBERSTATUSFLAGS: 347,
        MEMBERS: 286,
        MINACTUALVALUE: 383,
        MINPRESVALUE: 69,
        MINIMUMOFFTIME: 66,
        MINIMUMONTIME: 67,
        MINIMUMOUTPUT: 68,
        MINIMUMVALUE: 136,
        MINIMUMVALUETIMESTAMP: 150,
        MODE: 160,
        MODEL_NAME: 70,
        MODIFICATIONDATE: 71,
        MUSTERPOINT: 287,
        NEGATIVEACCESSRULES: 288,
        NETWORKACCESSSECURITYPOLICIES: 332,
        NETWORKINTERFACENAME: 424,
        NETWORKNUMBER: 425,
        NETWORKNUMBERQUALITY: 426,
        NETWORKTYPE: 427,
        NEXTSTOPPINGFLOOR: 476,
        NODESUBTYPE: 207,
        NODETYPE: 208,
        NOTIFICATIONCLASS: 17,
        NOTIFICATIONTHRESHOLD: 137,
        NOTIFYTYPE: 72,
        NUMBEROFAPDURETRIES: 73,
        NUMBEROFAUTHENTICATIONPOLICIES: 289,
        NUMBEROFSTATES: 74,
        OBJECTIDENTIFIER: 75,
        OBJECTLIST: 76,
        OBJECT_NAME: 77,
        OBJECTPROPERTYREFERENCE: 78,
        OBJECTTYPE: 79,
        OCCUPANCYCOUNT: 290,
        OCCUPANCYCOUNTADJUST: 291,
        OCCUPANCYCOUNTENABLE: 292,
        OCCUPANCYLOWERLIMIT: 294,
        OCCUPANCYLOWERLIMITENFORCED: 295,
        OCCUPANCYSTATE: 296,
        OCCUPANCYUPPERLIMIT: 297,
        OCCUPANCYUPPERLIMITENFORCED: 298,
        OPERATIONDIRECTION: 477,
        OPERATIONEXPECTED: 161,
        OPTIONAL: 80,
        OUT_OF_SERVICE: 81,
        OUTPUTUNITS: 82,
        PACKETREORDERTIME: 333,
        PASSBACKMODE: 300,
        PASSBACKTIMEOUT: 301,
        PASSENGERALARM: 478,
        POLARITY: 84,
        PORTFILTER: 363,
        POSITIVEACCESSRULES: 302,
        POWER: 384,
        POWERMODE: 479,
        PRESCALE: 185,
        PRESENT_VALUE: 85,
        PRIORITY: 86,
        PRIORITY_ARRAY: 87,
        PRIORITYFORWRITING: 88,
        PROCESSIDENTIFIER: 89,
        PROCESSIDENTIFIERFILTER: 361,
        PROFILELOCATION: 485,
        PROFILENAME: 168,
        PROGRAMCHANGE: 90,
        PROGRAMLOCATION: 91,
        PROGRAMSTATE: 92,
        PROPERTYLIST: 371,
        PROPORTIONALCONSTANT: 93,
        PROPORTIONALCONSTANTUNITS: 94,
        PROTOCOLLEVEL: 482,
        PROTOCOLOBJECTTYPESSUPPORTED: 96,
        PROTOCOLREVISION: 139,
        PROTOCOLSERVICESSUPPORTED: 97,
        PROTOCOLVERSION: 98,
        PULSERATE: 186,
        READONLY: 99,
        REASONFORDISABLE: 303,
        REASONFORHALT: 100,
        RECIPIENTLIST: 102,
        RECORDCOUNT: 141,
        RECORDSSINCENOTIFICATION: 140,
        REFERENCEPORT: 483,
        REGISTEREDCARCALL: 480,
        RELIABILITY: 103,
        RELIABILITYEVALUATIONINHIBIT: 357,
        RELINQUISHDEFAULT: 104,
        REPRESENTS: 491,
        REQUESTEDSHEDLEVEL: 218,
        REQUESTEDUPDATEINTERVAL: 348,
        REQUIRED: 105,
        RESOLUTION: 106,
        RESTARTNOTIFICATIONRECIPIENTS: 202,
        RESTORECOMPLETIONTIME: 340,
        RESTOREPREPARATIONTIME: 341,
        ROUTINGTABLE: 428,
        SCALE: 187,
        SCALEFACTOR: 188,
        SCHEDULEDEFAULT: 174,
        SECUREDSTATUS: 235,
        SECURITYPDUTIMEOUT: 334,
        SECURITYTIMEWINDOW: 335,
        SEGMENTATIONSUPPORTED: 107,
        SERIALNUMBER: 372,
        SETPOINT: 108,
        SETPOINTREFERENCE: 109,
        SETTING: 162,
        SHEDDURATION: 219,
        SHEDLEVELDESCRIPTIONS: 220,
        SHEDLEVELS: 221,
        SILENCED: 163,
        SLAVEADDRESSBINDING: 171,
        SLAVEPROXYENABLE: 172,
        STARTTIME: 142,
        STATECHANGEVALUES: 396,
        STATEDESCRIPTION: 222,
        STATETEXT: 110,
        STATUSFLAGS: 111,
        STOPTIME: 143,
        STOPWHENFULL: 144,
        STRIKECOUNT: 391,
        STRUCTUREDOBJECTLIST: 209,
        SUBORDINATEANNOTATIONS: 210,
        SUBORDINATELIST: 211,
        SUBORDINATENODETYPES: 487,
        SUBORDINATERELATIONSHIPS: 489,
        SUBORDINATETAGS: 488,
        SUBSCRIBEDRECIPIENTS: 362,
        SUPPORTEDFORMATCLASSES: 305,
        SUPPORTEDFORMATS: 304,
        SUPPORTEDSECURITYALGORITHMS: 336,
        SYSTEMSTATUS: 112,
        TAGS: 486,
        THREATAUTHORITY: 306,
        THREATLEVEL: 307,
        TIMEDELAY: 113,
        TIMEDELAYNORMAL: 356,
        TIMEOFACTIVETIMERESET: 114,
        TIMEOFDEVICERESTART: 203,
        TIMEOFSTATECOUNTRESET: 115,
        TIMEOFSTRIKECOUNTRESET: 392,
        TIMESYNCHRONIZATIONINTERVAL: 204,
        TIMESYNCHRONIZATIONRECIPIENTS: 116,
        TIMERRUNNING: 397,
        TIMERSTATE: 398,
        TOTALRECORDCOUNT: 145,
        TRACEFLAG: 308,
        TRACKINGVALUE: 164,
        TRANSACTIONNOTIFICATIONCLASS: 309,
        TRANSITION: 385,
        TRIGGER: 205,
        UNITS: 117,
        UPDATEINTERVAL: 118,
        UPDATEKEYSETTIMEOUT: 337,
        UPDATETIME: 189,
        USEREXTERNALIDENTIFIER: 310,
        USERINFORMATIONREFERENCE: 311,
        USERNAME: 317,
        USERTYPE: 318,
        USESREMAINING: 319,
        UTC_OFFSET: 119,
        UTCTIMESYNCHRONIZATIONRECIPIENTS: 206,
        VALIDSAMPLES: 146,
        VALUEBEFORECHANGE: 190,
        VALUECHANGETIME: 192,
        VALUESET: 191,
        VALUESOURCE: 433,
        VALUESOURCEARRAY: 434,
        VARIANCEVALUE: 151,
        VENDOR_IDENTIFIER: 120,
        VENDOR_NAME: 121,
        VERIFICATIONTIME: 326,
        VIRTUALMACADDRESSTABLE: 429,
        VTCLASSESSUPPORTED: 122,
        WEEKLYSCHEDULE: 123,
        WINDOWINTERVAL: 147,
        WINDOWSAMPLES: 148,
        WRITESTATUS: 370,
        ZONEFROM: 320,
        ZONEMEMBERS: 165,
        ZONETO: 32
    },

    ENGINEERING_UNITS: {
        // Full list in BACnetEngineeringUnits.h        
        METERSPERSECONDPERSECOND: 166,
        SQUAREMETERS: 0,
        SQUARECENTIMETERS: 116,
        SQUAREFEET: 1,
        SQUAREINCHES: 115,
        CURRENCY1: 105,
        CURRENCY2: 106,
        CURRENCY3: 107,
        CURRENCY4: 108,
        CURRENCY5: 109,
        CURRENCY6: 110,
        CURRENCY7: 111,
        CURRENCY8: 112,
        CURRENCY9: 113,
        CURRENCY10: 114,
        MILLIAMPERES: 2,
        AMPERES: 3,
        AMPERESPERMETER: 167,
        AMPERESPERSQUAREMETER: 168,
        AMPERESQUAREMETERS: 169,
        DECIBELS: 199,
        DECIBELSMILLIVOLT: 200,
        DECIBELSVOLT: 201,
        FARADS: 170,
        HENRYS: 171,
        OHMS: 4,
        OHMMETERSQUAREDPERMETER: 237,
        OHMMETERS: 172,
        MILLIOHMS: 145,
        KILOHMS: 122,
        MEGOHMS: 123,
        MICROSIEMENS: 190,
        MILLISIEMENS: 202,
        SIEMENS: 173,
        SIEMENSPERMETER: 174,
        TESLAS: 175,
        VOLTS: 5,
        MILLIVOLTS: 124,
        KILOVOLTS: 6,
        MEGAVOLTS: 7,
        VOLTAMPERES: 8,
        KILOVOLTAMPERES: 9,
        MEGAVOLTAMPERES: 10,
        VOLT_AMPERES_REACTIVE: 11,
        KILOVOLTAMPERESREACTIVE: 12,
        MEGAVOLTAMPERESREACTIVE: 13,
        VOLTSPERDEGREEKELVIN: 176,
        VOLTSPERMETER: 177,
        DEGREESPHASE: 14,
        POWERFACTOR: 15,
        WEBERS: 178,
        AMPERESECONDS: 238,
        VOLTAMPEREHOURS: 239,
        KILOVOLTAMPEREHOURS: 240,
        MEGAVOLTAMPEREHOURS: 241,
        VOLTAMPEREHOURSREACTIVE: 242,
        KILOVOLTAMPEREHOURSREACTIVE: 243,
        MEGAVOLTAMPEREHOURSREACTIVE: 244,
        VOLTSQUAREHOURS: 245,
        AMPERESQUAREHOURS: 246,
        JOULES: 16,
        KILOJOULES: 17,
        KILOJOULESPERKILOGRAM: 125,
        MEGAJOULES: 126,
        WATT_HOURS: 18,
        KILOWATT_HOURS: 19,
        MEGAWATT_HOURS: 146,
        WATTHOURSREACTIVE: 203,
        KILOWATTHOURSREACTIVE: 204,
        MEGAWATTHOURSREACTIVE: 205,
        BTUS: 20,
        KILOBTUS: 147,
        MEGABTUS: 148,
        THERMS: 21,
        TONHOURS: 22,
        JOULESPERKILOGRAMDRYAIR: 23,
        KILOJOULESPERKILOGRAMDRYAIR: 149,
        MEGAJOULESPERKILOGRAMDRYAIR: 150,
        BTUSPERPOUNDDRYAIR: 24,
        BTUSPERPOUND: 117,
        JOULESPERDEGREEKELVIN: 127,
        KILOJOULESPERDEGREEKELVIN: 151,
        MEGAJOULESPERDEGREEKELVIN: 152,
        JOULESPERKILOGRAMDEGREEKELVIN: 128,
        NEWTON: 153,
        CYCLESPERHOUR: 25,
        CYCLESPERMINUTE: 26,
        HERTZ: 27,
        KILOHERTZ: 129,
        MEGAHERTZ: 130,
        PERHOUR: 131,
        GRAMSOFWATERPERKILOGRAMDRYAIR: 28,
        PERCENTRELATIVEHUMIDITY: 29,
        MICROMETERS: 194,
        MILLIMETERS: 30,
        CENTIMETERS: 118,
        KILOMETERS: 193,
        METERS: 31,
        INCHES: 32,
        FEET: 33,
        CANDELAS: 179,
        CANDELASPERSQUAREMETER: 180,
        WATTSPERSQUAREFOOT: 34,
        WATTSPERSQUAREMETER: 35,
        LUMENS: 36,
        LUXES: 37,
        FOOTCANDLES: 38,
        MILLIGRAMS: 196,
        GRAMS: 195,
        KILOGRAMS: 39,
        POUNDSMASS: 40,
        TONS: 41,
        GRAMSPERSECOND: 154,
        GRAMSPERMINUTE: 155,
        KILOGRAMSPERSECOND: 42,
        KILOGRAMSPERMINUTE: 43,
        KILOGRAMSPERHOUR: 44,
        POUNDSMASSPERSECOND: 119,
        POUNDSMASSPERMINUTE: 45,
        POUNDSMASSPERHOUR: 46,
        TONSPERHOUR: 156,
        MILLIWATTS: 132,
        WATTS: 47,
        KILOWATTS: 48,
        MEGAWATTS: 49,
        BTUSPERHOUR: 50,
        KILOBTUSPERHOUR: 157,
        JOULEPERHOURS: 247,
        HORSEPOWER: 51,
        TONSREFRIGERATION: 52,
        PASCALS: 53,
        HECTOPASCALS: 133,
        KILOPASCALS: 54,
        MILLIBARS: 134,
        BARS: 55,
        POUNDSFORCEPERSQUAREINCH: 56,
        MILLIMETERSOFWATER: 206,
        CENTIMETERSOFWATER: 57,
        INCHESOFWATER: 58,
        MILLIMETERSOFMERCURY: 59,
        CENTIMETERSOFMERCURY: 60,
        INCHESOFMERCURY: 61,
        DEGREESCELSIUS: 62,
        DEGREESKELVIN: 63,
        DEGREESKELVINPERHOUR: 181,
        DEGREESKELVINPERMINUTE: 182,
        DEGREESFAHRENHEIT: 64,
        DEGREEDAYSCELSIUS: 65,
        DEGREEDAYSFAHRENHEIT: 66,
        DELTADEGREESFAHRENHEIT: 120,
        DELTADEGREESKELVIN: 121,
        YEARS: 67,
        MONTHS: 68,
        WEEKS: 69,
        DAYS: 70,
        HOURS: 71,
        MINUTES: 72,
        SECONDS: 73,
        HUNDREDTHSSECONDS: 158,
        MILLISECONDS: 159,
        NEWTONMETERS: 160,
        MILLIMETERSPERSECOND: 161,
        MILLIMETERSPERMINUTE: 162,
        METERSPERSECOND: 74,
        METERSPERMINUTE: 163,
        METERSPERHOUR: 164,
        KILOMETERSPERHOUR: 75,
        FEETPERSECOND: 76,
        FEETPERMINUTE: 77,
        MILESPERHOUR: 78,
        CUBICFEET: 79,
        CUBICMETERS: 80,
        IMPERIALGALLONS: 81,
        MILLILITERS: 197,
        LITERS: 82,
        USGALLONS: 83,
        CUBICFEETPERSECOND: 142,
        CUBICFEETPERMINUTE: 84,
        MILLIONSTANDARDCUBICFEETPERMINUTE: 254,
        CUBICFEETPERHOUR: 191,
        CUBICFEETPERDAY: 248,
        STANDARDCUBICFEETPERDAY: 47808,
        MILLIONSTANDARDCUBICFEETPERDAY: 47809,
        THOUSANDCUBICFEETPERDAY: 47810,
        THOUSANDSTANDARDCUBICFEETPERDAY: 47811,
        POUNDSMASSPERDAY: 47812,
        CUBICMETERSPERSECOND: 85,
        CUBICMETERSPERMINUTE: 165,
        CUBICMETERSPERHOUR: 135,
        CUBICMETERSPERDAY: 249,
        IMPERIALGALLONSPERMINUTE: 86,
        MILLILITERSPERSECOND: 198,
        LITERSPERSECOND: 87,
        LITERSPERMINUTE: 88,
        LITERSPERHOUR: 136,
        USGALLONSPERMINUTE: 89,
        USGALLONSPERHOUR: 192,
        DEGREESANGULAR: 90,
        DEGREESCELSIUSPERHOUR: 91,
        DEGREESCELSIUSPERMINUTE: 92,
        DEGREESFAHRENHEITPERHOUR: 93,
        DEGREESFAHRENHEITPERMINUTE: 94,
        JOULESECONDS: 183,
        KILOGRAMSPERCUBICMETER: 186,
        KILOWATTHOURSPERSQUAREMETER: 137,
        KILOWATTHOURSPERSQUAREFOOT: 138,
        WATTHOURSPERCUBICMETER: 250,
        JOULESPERCUBICMETER: 251,
        MEGAJOULESPERSQUAREMETER: 139,
        MEGAJOULESPERSQUAREFOOT: 140,
        MOLEPERCENT: 252,
        NO_UNITS: 95,
        NEWTONSECONDS: 187,
        NEWTONSPERMETER: 188,
        PARTSPERMILLION: 96,
        PARTSPERBILLION: 97,
        PASCALSECONDS: 253,
        PERCENT: 98,
        PERCENTOBSCURATIONPERFOOT: 143,
        PERCENTOBSCURATIONPERMETER: 144,
        PERCENTPERSECOND: 99,
        PERMINUTE: 100,
        PERSECOND: 101,
        PSIPERDEGREEFAHRENHEIT: 102,
        RADIANS: 103,
        RADIANSPERSECOND: 184,
        REVOLUTIONSPERMINUTE: 104,
        SQUAREMETERSPERNEWTON: 185,
        WATTSPERMETERPERDEGREEKELVIN: 189,
        WATTSPERSQUAREMETERDEGREEKELVIN: 141,
        PERMILLE: 207,
        GRAMSPERGRAM: 208,
        KILOGRAMSPERKILOGRAM: 209,
        GRAMSPERKILOGRAM: 210,
        MILLIGRAMSPERGRAM: 211,
        MILLIGRAMSPERKILOGRAM: 212,
        GRAMSPERMILLILITER: 213,
        GRAMSPERLITER: 214,
        MILLIGRAMSPERLITER: 215,
        MICROGRAMSPERLITER: 216,
        GRAMSPERCUBICMETER: 217,
        MILLIGRAMSPERCUBICMETER: 218,
        MICROGRAMSPERCUBICMETER: 219,
        NANOGRAMSPERCUBICMETER: 220,
        GRAMSPERCUBICCENTIMETER: 221,
        BECQUERELS: 222,
        KILOBECQUERELS: 223,
        MEGABECQUERELS: 224,
        GRAY: 225,
        MILLIGRAY: 226,
        MICROGRAY: 227,
        SIEVERTS: 228,
        MILLISIEVERTS: 229,
        MICROSIEVERTS: 230,
        MICROSIEVERTSPERHOUR: 231,
        MILLIREMS: 47814,
        MILLIREMSPERHOUR: 47815,
        DECIBELSA: 232,
        NEPHELOMETRICTURBIDITYUNIT: 233,
        PH: 234,
        GRAMSPERSQUAREMETER: 235,
        MINUTESPERDEGREEKELVIN: 236
    }




};
